/*
 * yasapi_strategy.c
 * Copyright (C) 2015-2016 Peter Belkner <pbelkner@snafu.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#include <yasapi.h>

///////////////////////////////////////////////////////////////////////////////
static HRESULT InitializePush(Player *pPlayer, DWORD dwStreamFlags,
    REFERENCE_TIME hnsDevicePeriod)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->options.common.nDebug;
#endif // }
  AUDCLNT_SHAREMODE eShareMode=pPlayer->open.eShareMode;
  IAudioClient *pClient=pPlayer->open.pClient;
  WAVEFORMATEX *pwfx=&pPlayer->open.wfxx.Format;

  DPRINTF(debug,0,"  intitializing audio client %s/PUSH with %I64d/0 hns\n",
      YASAPI_SHAREMODE_LABEL(eShareMode),hnsDevicePeriod);

  return pClient->lpVtbl->Initialize(pClient,
    eShareMode,           // [in]  AUDCLNT_SHAREMODE ShareMode,
    dwStreamFlags,        // [in]  DWORD StreamFlags,
    hnsDevicePeriod,      // [in]  REFERENCE_TIME hnsBufferDuration,
    0,                    // [in]  REFERENCE_TIME hnsPeriodicity,
    pwfx,                 // [in]  const WAVEFORMATEX *pFormat,
    NULL                  // [in]  LPCGUID AudioPlayerGuid
  );
}

#if 0 // {
static void BalanceStartPush(Player *pPlayer)
{
  Channel *pChannel=pPlayer->pChannel;
  int bBalance=pChannel->options.bBalance;

  pPlayer->balance.q=bBalance?pChannel->options.qBalanceStart:0.5;
}

static void BalanceIteratePush(Player *pPlayer, UINT32 uFramesWrite)
{
  Channel *pChannel=pPlayer->pChannel;
  int bBalance=pChannel->options.bBalance;
  double q;

  if (bBalance&&0ull<uFramesWrite) {
    q=0.5*pPlayer->balance.q*pPlayer->uFrames/uFramesWrite;
    pPlayer->balance.q
        +=(q-pPlayer->balance.q)/(++pPlayer->balance.count);
  }
}
#endif // }

static LONGLONG IntervalPush(Player *pPlayer, LONGLONG time)
{
  return MulDiv(time,1,2);
}

static void SetTimerPush(Player *pPlayer, LONGLONG time, BOOL bFlush)
{
  //PlayerSetTimer(pPlayer,0.5*time+0.5);
  PlayerSetTimer(pPlayer,IntervalPush(pPlayer,time));
}

static BOOL NeedDevicePeriodPush(Player *pPlayer)
{
  return TRUE;
}

static BOOL NeedPaddingPush(Player *pPlayer)
{
  return TRUE;
}

#if 0 // {
static BOOL CanAdaptPush(Player *pPlayer)
{
  return TRUE;
}

static BOOL IsPullPush(Player *pPlayer)
{
  return FALSE;
}
#endif // }

static HRESULT SetEventPush(Player *pPlayer)
{
  return S_OK;
}

static HANDLE GetEventPush(Player *pPlayer)
{
  return NULL;
}

static UINT32 GetFramesPaddingPush(Player *pPlayer)
{
  return PlayerGetFramesPadding(pPlayer);
}

const Strategy gcStrategyPush={
  InitializePush,
#if 0 // {
  BalanceStartPush,
  BalanceIteratePush,
#endif // }
  IntervalPush,
  SetTimerPush,
  NeedDevicePeriodPush,
  NeedPaddingPush,
#if 0 // {
  CanAdaptPush,
  IsPullPush,
#endif // }
  SetEventPush,
  GetEventPush,
  GetFramesPaddingPush
};

///////////////////////////////////////////////////////////////////////////////
static HRESULT InitializePull(Player *pPlayer, DWORD dwStreamFlags,
    REFERENCE_TIME hnsDevicePeriod)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->options.common.nDebug;
#endif // }
  AUDCLNT_SHAREMODE eShareMode=pPlayer->open.eShareMode;
  IAudioClient *pClient=pPlayer->open.pClient;
  WAVEFORMATEX *pwfx=&pPlayer->open.wfxx.Format;
  HRESULT hr;

  dwStreamFlags|=AUDCLNT_STREAMFLAGS_EVENTCALLBACK;

  switch (eShareMode) {
  case AUDCLNT_SHAREMODE_SHARED:
    // For a shared-mode stream that uses event-driven buffering, the
    // caller must set both hnsPeriodicity and hnsBufferDuration to 0. The
    // Initialize method determines how large a buffer to allocate based on
    // the scheduling period of the audio engine. Although the client's buffer
    // processing thread is event driven, the basic buffer management process,
    // as described previously, is unaltered. Each time the thread awakens,
    // it should call IAudioClient::GetCurrentPadding to determine how much
    // data to write to a rendering buffer or read from a capture buffer. In
    // contrast to the two buffers that the Initialize method allocates for an
    // exclusive-mode stream that uses event-driven buffering, a shared-mode
    // stream requires a single buffer.
    DPUTS(debug,0,"  intitializing audio client SHARED/PULL with 0/0 hns\n");

    return pClient->lpVtbl->Initialize(pClient,
      eShareMode,         // [in]  AUDCLNT_SHAREMODE ShareMode,
      dwStreamFlags,      // [in]  DWORD StreamFlags,
      0,                  // [in]  REFERENCE_TIME hnsBufferDuration,
      0,                  // [in]  REFERENCE_TIME hnsPeriodicity,
      pwfx,               // [in]  const WAVEFORMATEX *pFormat,
      NULL                // [in]  LPCGUID AudioPlayerGuid
    );
  case AUDCLNT_SHAREMODE_EXCLUSIVE:
    // For an exclusive-mode stream that uses event-driven buffering,
    // the caller must specify nonzero values for hnsPeriodicity
    // and hnsBufferDuration, and the values of these two parameters must
    // be equal. The Initialize method allocates two buffers for the stream.
    // Each buffer is equal in duration to the value of the hnsBufferDuration
    // parameter. Following the Initialize call for a rendering stream, the
    // caller should fill the first of the two buffers before starting the
    // stream. For a capture stream, the buffers are initially empty, and
    // the caller should assume that each buffer remains empty until the
    // event for that buffer is signaled. While the stream is running, the
    // system alternately sends one buffer or the other to the client—this
    // form of double buffering is referred to as "ping-ponging". Each time
    // the client receives a buffer from the system (which the system
    // indicates by signaling the event), the client must process the entire
    // buffer. For example, if the client requests a packet size from the
    // IAudioRenderClient::GetBuffer method that does not match the buffer
    // size, the method fails. Calls to the IAudioClient::GetCurrentPadding
    // method are unnecessary because the packet size must always equal the
    // buffer size. In contrast to the buffering modes discussed previously,
    // the latency for an event-driven, exclusive-mode stream depends directly
    // on the buffer size.
    DPRINTF(debug,0,"  intitializing audio client EXCLUSIVE/PULL"
        " with %I64d/%I64d hns\n",hnsDevicePeriod,hnsDevicePeriod);

    return pClient->lpVtbl->Initialize(pClient,
      eShareMode,         // [in]  AUDCLNT_SHAREMODE ShareMode,
      dwStreamFlags,      // [in]  DWORD StreamFlags,
      hnsDevicePeriod,    // [in]  REFERENCE_TIME hnsBufferDuration,
      hnsDevicePeriod,    // [in]  REFERENCE_TIME hnsPeriodicity,
      pwfx,               // [in]  const WAVEFORMATEX *pFormat,
      NULL                // [in]  LPCGUID AudioPlayerGuid
    );
  default:
    return E_FAIL;
  }
}

#if 0 // {
static void BalanceStartPull(Player *pPlayer)
{
  switch (pPlayer->eShareMode) {
  case AUDCLNT_SHAREMODE_SHARED:
    pPlayer->balance.q=0.5;
    break;
  case AUDCLNT_SHAREMODE_EXCLUSIVE:
    pPlayer->balance.q=1.0;
    break;
  default:
    break;
  }
}

static void BalanceIteratePull(Player *pPlayer, UINT32 uFramesWrite)
{
}
#endif // }

static LONGLONG IntervalPull(Player *pPlayer, LONGLONG time)
{
    switch (pPlayer->open.eShareMode) {
    case AUDCLNT_SHAREMODE_SHARED:
      return MulDiv(time,1,2);
    case AUDCLNT_SHAREMODE_EXCLUSIVE:
      return time;
    default:
      return MulDiv(time,1,2);
    }
}

static void SetTimerPull(Player *pPlayer, LONGLONG time, BOOL bFlush)
{
  if (bFlush) {
#if 0 // {
    switch (pPlayer->open.eShareMode) {
    case AUDCLNT_SHAREMODE_SHARED:
      PlayerSetTimer(pPlayer,0.5*time+0.5);
      break;
    case AUDCLNT_SHAREMODE_EXCLUSIVE:
      PlayerSetTimer(pPlayer,time+0.5);
      break;
    default:
      PlayerSetTimer(pPlayer,0.5*time+0.5);
      break;
    }
#elif 0 // } {
    PlayerSetTimer(pPlayer,IntervalPull(pPlayer,time));
#endif // }
  }
}

static BOOL NeedDevicePeriodPull(Player *pPlayer)
{
  switch (pPlayer->open.eShareMode) {
  case AUDCLNT_SHAREMODE_SHARED:
    return FALSE;
  case AUDCLNT_SHAREMODE_EXCLUSIVE:
    return TRUE;
  default:
    return FALSE;
  }
}

static BOOL NeedPaddingPull(Player *pPlayer)
{
  switch (pPlayer->open.eShareMode) {
  case AUDCLNT_SHAREMODE_SHARED:
    return FALSE;
  case AUDCLNT_SHAREMODE_EXCLUSIVE:
    return TRUE;
  default:
    return FALSE;
  }
}

#if 0 // {
static BOOL CanAdaptPull(Player *pPlayer)
{
  switch (pPlayer->eShareMode) {
  case AUDCLNT_SHAREMODE_SHARED:
    return TRUE;
  case AUDCLNT_SHAREMODE_EXCLUSIVE:
    return FALSE;
  default:
    return TRUE;
  }
}

static BOOL IsPullPull(Player *pPlayer)
{
  return TRUE;
}
#endif // }

static HRESULT SetEventPull(Player *pPlayer)
{
#if defined (YASAPI_DEBUG) // {
  int debug=pPlayer->options.common.nDebug;
#endif // }
  IAudioClient *pClient=pPlayer->open.pClient;
  HRESULT hr;

  hr=pClient->lpVtbl->SetEventHandle(pClient,
    pPlayer->base.hEvent    // [in] HANDLE eventHandle
  );

  if (!FAILED(hr))
    DPUTS(debug,0,"  event handle set\n");

  return hr;
}

static HANDLE GetEventPull(Player *pPlayer)
{
  return pPlayer->state<PLAYER_STATE_PLAY?NULL:pPlayer->base.hEvent;
}

static UINT32 GetFramesPaddingPull(Player *pPlayer)
{
  switch (pPlayer->open.eShareMode) {
  case AUDCLNT_SHAREMODE_SHARED:
    return PlayerGetFramesPadding(pPlayer);
  case AUDCLNT_SHAREMODE_EXCLUSIVE:
    return 0;
  default:
    return PlayerGetFramesPadding(pPlayer);
  }
}

const Strategy gcStrategyPull={
  InitializePull,
#if 0 // {
  BalanceStartPull,
  BalanceIteratePull,
#endif // }
  IntervalPull,
  SetTimerPull,
  NeedDevicePeriodPull,
  NeedPaddingPull,
#if 0 // {
  CanAdaptPull,
  IsPullPull,
#endif // }
  SetEventPull,
  GetEventPull,
  GetFramesPaddingPull
};
